/*
 * libopenraw - dumputils.hpp
 *
 * Copyright (C) 2020 Hubert Figuière
 *
 * This library is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <iostream>
#include <string>

#include <libopenraw/libopenraw.h>

/** @defgroup tools Tools
 * @brief Tools for libopenraw
 * @{
 */

/** @brief Return a string for the raw file type */
std::string typeToString(or_rawfile_type t);
/** @brief Dump raw file info */
void dump_file_info(std::ostream& out, ORRawFileRef rf, bool dev_mode);

/** @} */
