// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::{IOStream, InputStream, OutputStream, ffi};
use glib::{prelude::*, translate::*};

glib::wrapper! {
    #[doc(alias = "GSimpleIOStream")]
    pub struct SimpleIOStream(Object<ffi::GSimpleIOStream>) @extends IOStream;

    match fn {
        type_ => || ffi::g_simple_io_stream_get_type(),
    }
}

impl SimpleIOStream {
    #[doc(alias = "g_simple_io_stream_new")]
    pub fn new(
        input_stream: &impl IsA<InputStream>,
        output_stream: &impl IsA<OutputStream>,
    ) -> SimpleIOStream {
        unsafe {
            IOStream::from_glib_full(ffi::g_simple_io_stream_new(
                input_stream.as_ref().to_glib_none().0,
                output_stream.as_ref().to_glib_none().0,
            ))
            .unsafe_cast()
        }
    }
}
