use crate::geometry::Translation;

/// A 1-dimensional translation.
pub type Translation1<T> = Translation<T, 1>;

/// A 2-dimensional translation.
pub type Translation2<T> = Translation<T, 2>;

/// A 3-dimensional translation.
pub type Translation3<T> = Translation<T, 3>;

/// A 4-dimensional translation.
pub type Translation4<T> = Translation<T, 4>;

/// A 5-dimensional translation.
pub type Translation5<T> = Translation<T, 5>;

/// A 6-dimensional translation.
pub type Translation6<T> = Translation<T, 6>;
