// This file was generated by gir (https://github.com/gtk-rs/gir)
// from gir-files (https://github.com/gtk-rs/gir-files)
// DO NOT EDIT

use crate::ffi;
use glib::translate::*;

#[doc(alias = "g_unix_is_mount_path_system_internal")]
pub fn is_mount_path_system_internal(mount_path: impl AsRef<std::path::Path>) -> bool {
    unsafe {
        from_glib(ffi::g_unix_is_mount_path_system_internal(
            mount_path.as_ref().to_glib_none().0,
        ))
    }
}

#[doc(alias = "g_unix_is_system_device_path")]
pub fn is_system_device_path(device_path: &str) -> bool {
    unsafe {
        from_glib(ffi::g_unix_is_system_device_path(
            device_path.to_glib_none().0,
        ))
    }
}

#[doc(alias = "g_unix_is_system_fs_type")]
pub fn is_system_fs_type(fs_type: &str) -> bool {
    unsafe { from_glib(ffi::g_unix_is_system_fs_type(fs_type.to_glib_none().0)) }
}
